# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

notifications   = True
webcam          = True
clipboard       = True
audio           = True
av_sync         = True
fileprint       = True
mmap            = True
input_devices   = True
commands        = True
dbus            = True
encoding        = True
logging         = True
network_state   = True
shell           = False
display         = True
windows         = True
rfb             = True
