BUILD_BIT='64bit'
BUILD_CPU='x86_64'
BUILD_DATE='2021-10-04'
BUILD_MACHINE='x86_64'
BUILD_OS='Linux Fedora 34 ThirtyFour'
BUILD_TIME='16:22'
BUILT_BY='antoine'
BUILT_ON='localhost.localdomain'
COMPILER_VERSION='gcc (GCC) 11.2.1 20210728 (Red Hat 11.2.1-1)'
CYTHON_VERSION='0.29.21'
LINKER_VERSION='GNU ld version 2.35.2-6.fc34'
PYTHON_VERSION='3.9.7'
RELEASE_BUILD=True
lib_gobject_introspection='1.68.0'
lib_gtk='3.24.30'
lib_gtk_x11='3.24.30'
lib_nvenc='10'
lib_nvfbc='6.0'
lib_py3cairo='1.20.0'
lib_pygobject='3.40.1'
lib_python3='3.9'
lib_vpx='1.9.0'
lib_x11='1.7.2'
lib_x265='3.5'
lib_xcomposite='0.4.5'
lib_xdamage='1.1.5'
lib_xext='1.3.4'
lib_xfixes='6.0.0'
lib_xkbfile='1.1.0'
lib_xrandr='1.5.2'
lib_xtst='1.2.3'
